""" opentrons.protocols: functions and state for handling, executing, and
simulating protocols.

This genericizes over v1 and v2 protocols and allows for version inference
and version-independent apis for simulating and executing protocols.
"""
